#!/usr/bin/env python

"""
Inspect source files, obtaining details of classes and attributes.

Copyright (C) 2007 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import compiler.ast
from compiler.visitor import ASTVisitor
try:
    set
except NameError:
    from sets import Set as set

class InspectError(Exception):

    "An inspection error."

    pass

# Mix-ins and abstract classes.

class NamespaceDict:

    "A mix-in providing dictionary methods."

    def __init__(self):
        self.namespace = {}
        self.globals = set()

    def __getitem__(self, name):
        return self.namespace[name]

    def __setitem__(self, name, value):
        if name not in self.globals:
            self.namespace[name] = value

    def __delitem__(self, name):
        del self.namespace[name]

    def has_key(self, name):
        return self.namespace.has_key(name)

    def keys(self):
        return self.namespace.keys()

    def values(self):
        return self.namespace.values()

    def items(self):
        return self.namespace.items()

    def make_global(self, name):
        if not self.namespace.has_key(name):
            self.globals.add(name)
        else:
            raise InspectError, "Name %r is global and local in %r" % (name, self)

class Naming:

    "A mix-in providing naming conveniences."

    def full_name(self):
        if self.name is not None:
            return self.parent_name + "." + self.name
        else:
            return self.parent_name

# Program data structures.

class Attr:

    "An attribute entry."

    def __init__(self, position, parent, value=None):
        self.position = position
        self.parent = parent
        self.value = value

    def __repr__(self):
        return "Attr(%d, %r, %r)" % (self.position, self.parent, self.value)

class Class(NamespaceDict, Naming):

    "An inspected class."

    def __init__(self, name, parent_name):
        NamespaceDict.__init__(self)
        self.name = name
        self.parent_name = parent_name

        self.bases = []
        self.instattr = set() # instance attributes

        self.all_instattr = None # cache for instance_attributes
        self.all_instattr_names = None # from all_instattr
        self.classattr = None # cache for class_attributes
        self.classattr_names = None # from classattr
        self.all_classattr = None # cache for all_class_attributes
        self.all_classattr_names = None # from all_classattr
        self.allattr = None # cache for all_attributes
        self.allattr_names = None # from allattr

        # Image generation details.

        self.location = None

    def __repr__(self):
        return "Class(%r, %r, location=%r)" % (self.name, self.parent_name, self.location)

    def add_base(self, base):
        self.bases.append(base)

    def add_instance_attribute(self, name):
        self.instattr.add(name)

    def class_attribute_names(self):

        "Return the attribute names provided by this class only."

        if self.classattr_names is None:
            self.class_attributes()
        return self.classattr_names

    def class_attributes(self):

        "Return class attributes provided by this class only."

        if self.classattr is None:
            self.classattr = {}
            self.classattr_names = self.keys()

            for i, name in enumerate(self.classattr_names):
                self.classattr[name] = Attr(i, self, self[name])

        return self.classattr

    def all_class_attribute_names(self):

        "Return the attribute names provided by classes in this hierarchy."

        if self.all_classattr_names is None:
            self.all_class_attributes()
        return self.all_classattr_names

    def all_class_attributes(self):

        "Return all class attributes, indicating the class which provides them."

        if self.all_classattr is None:
            self.all_classattr = {}

            reversed_bases = self.bases[:]
            reversed_bases.reverse()
            for cls in reversed_bases:
                self.all_classattr.update(cls.all_class_attributes())

            # Record attributes provided by this class, along with their
            # positions.

            self.all_classattr.update(self.class_attributes())

        return self.all_classattr

    def instance_attribute_names(self):

        "Return the instance attribute names provided by the class."

        if self.all_instattr_names is None:
            self.instance_attributes()
        return self.all_instattr_names

    def instance_attributes(self):

        "Return instance-only attributes for instances of this class."

        if self.all_instattr is None:
            self.all_instattr = {}
            instattr = set()

            reversed_bases = self.bases[:]
            reversed_bases.reverse()
            for cls in reversed_bases:
                instattr.update(cls.instance_attributes().keys())

            # Record instance attributes provided by this class and its bases,
            # along with their positions.

            instattr.update(self.instattr)

            for i, name in enumerate(instattr):
                self.all_instattr[name] = Attr(i, None)

            self.all_instattr_names = self.all_instattr.keys()

        return self.all_instattr

    def all_attribute_names(self):

        """
        Return the names of all attributes provided by instances of this class.
        """

        self.allattr_names = self.allattr_names or self.all_attributes().keys()
        return self.allattr_names

    def all_attributes(self):

        """
        Return all attributes for an instance, indicating either the class which
        provides them or that the instance itself provides them.
        """

        if self.allattr is None:
            self.allattr = {}
            self.allattr.update(self.all_class_attributes())
            for i, name in enumerate(self.instance_attribute_names()):
                if self.allattr.has_key(name):
                    print "Instance attribute %r in %r overrides class attribute." % (name, self)
                self.allattr[name] = Attr(i, None)
        return self.allattr

class Function(NamespaceDict, Naming):

    "An inspected function."

    def __init__(self, name, parent_name, argnames, has_star, has_dstar):
        NamespaceDict.__init__(self)
        self.name = name
        self.parent_name = parent_name
        self.argnames = argnames
        self.has_star = has_star
        self.has_dstar = has_dstar

        self.localnames = None # cache for locals
        self.alllocalnames = None # cache for all_locals

        # Image generation details.

        self.location = None

    def __repr__(self):
        return "Function(%r, %r, %r, %r, %r, location=%r)" % (
            self.name, self.parent_name, self.argnames, self.has_star, self.has_dstar, self.location
            )

    def make_global(self, name):
        if name not in self.argnames and not self.has_key(name):
            self.globals.add(name)
        else:
            raise InspectError, "Name %r is global and local in %r" % (name, self)

    def parameters(self):

        """
        Return a dictionary mapping parameter names to their position in the
        parameter list.
        """

        parameters = {}
        for i, name in enumerate(self.argnames):
            parameters[name] = i
        return parameters

    def locals(self):

        "Return a dictionary mapping names to local details."

        if self.localnames is None:
            self.localnames = {}
            start = len(self.argnames)
            for i, name in enumerate(self.keys()):
                self.localnames[name] = Attr(start + i, None)
        return self.localnames

    def all_locals(self):

        "Return a dictionary mapping names to local and parameter details."

        if self.alllocalnames is None:
            self.alllocalnames = {}
            self.alllocalnames.update(self.locals())
            for i, name in enumerate(self.argnames):
                self.alllocalnames[name] = Attr(i, None)
        return self.alllocalnames

class UnresolvedName(NamespaceDict, Naming):

    "A module, class or function which was mentioned but could not be imported."

    def __init__(self, name, parent_name):
        NamespaceDict.__init__(self)
        self.name = name
        self.parent_name = parent_name

    def all_class_attributes(self):
        return {}

    def instance_attributes(self):
        return {}

    def __repr__(self):
        return "UnresolvedName(%r, %r)" % (self.name, self.parent_name)

# Program visitors.

class Module(ASTVisitor, NamespaceDict):

    "An inspected module."

    def __init__(self, name, importer=None):
        ASTVisitor.__init__(self)
        NamespaceDict.__init__(self)
        self.visitor = self

        self.name = name
        self.importer = importer
        self.loaded = 0

        # Module attributes.

        self.modattr = None # cache for module_attributes
        self.modattr_names = None # from modattr

        # Complete lists of classes and functions.

        self.all_objects = []

        # Current expression state.

        self.expr = None

        # Namespace state.

        self.in_init = 0
        self.namespaces = []
        self.module = None

        # Image generation details.

        self.location = None

    def parse(self, filename):

        "Parse the file having the given 'filename'."

        module = compiler.parseFile(filename)
        self.process(module)

    def process(self, module):

        "Process the given 'module'."

        self.module = module
        processed = self.dispatch(module)
        if self.has_key("__all__"):
            all = self["__all__"]
            if isinstance(all, compiler.ast.List):
                for n in all.nodes:
                    self[n.value] = self.importer.add_module(self.name + "." + n.value)
        return processed

    def vacuum(self):

        "Vacuum the module namespace, removing unloaded module references."

        for name, value in self.items():
            if isinstance(value, Module) and not value.loaded:
                del self[name]

    def full_name(self):
        return self.name

    def __repr__(self):
        return "Module(%r, location=%r)" % (self.name, self.location)

    # Attribute methods.

    def module_attribute_names(self):

        "Return the module attribute names provided by the module."

        if self.modattr_names is None:
            self.module_attributes()
        return self.modattr_names

    def module_attributes(self):

        "Return a dictionary mapping names to module attributes."

        if self.modattr is None:
            self.modattr = {}
            self.modattr_names = self.keys()
            for i, name in enumerate(self.modattr_names):
                self.modattr[name] = Attr(i, self, self[name])

        return self.modattr

    # Namespace methods.

    def store(self, name, obj):

        "Record attribute or local 'name', storing 'obj'."

        if not self.namespaces:
            self[name] = obj
        else:
            self.namespaces[-1][name] = obj

        # Record all non-local objects.

        if not (self.namespaces and isinstance(self.namespaces[-1], Function)):
            self.all_objects.append(obj)

    def store_attr(self, name):

        "Record instance attribute 'name' in the current class."

        if self.in_init:
            # Current namespace is the function.
            # Previous namespace is the class.
            self.namespaces[-2].add_instance_attribute(name)

    def get_parent(self):
        return (self.namespaces[-1:] or [self])[0]

    # Visitor methods.

    def default(self, node, *args):
        raise InspectError, node.__class__

    def dispatch(self, node, *args):
        return ASTVisitor.dispatch(self, node, *args)

    def NOP(self, node):
        return node

    visitAdd = NOP

    visitAnd = NOP

    visitAssert = NOP

    def visitAssign(self, node):
        self.expr = self.dispatch(node.expr)
        for n in node.nodes:
            self.dispatch(n)
        return None

    def visitAssAttr(self, node):
        expr = self.dispatch(node.expr)
        if expr is not None and isinstance(expr, Self):
            self.store_attr(node.attrname)
        return None

    def visitAssList(self, node):
        for n in node.nodes:
            self.dispatch(n)
        return None

    def visitAssName(self, node):
        self.store(node.name, self.expr)
        return None

    visitAssTuple = visitAssList

    visitAugAssign = NOP

    visitBackquote = NOP

    visitBitand = NOP

    visitBitor = NOP

    visitBitxor = NOP

    visitBreak = NOP

    visitCallFunc = NOP

    def visitClass(self, node):
        if self.namespaces:
            print "Class %r in %r is not global: ignored." % (node.name, self)
        else:
            cls = Class(node.name, self.get_parent().full_name())
            for base in node.bases:
                base_ref = self.dispatch(base)
                if base_ref is None:
                    raise InspectError, "Base class %r for class %r in %r is not found: it may be hidden in some way." % (base, self, cls)
                cls.add_base(base_ref)

            # Make an entry for the class.

            self.store(node.name, cls)

            self.namespaces.append(cls)
            self.dispatch(node.code)
            self.namespaces.pop()

        return node

    visitCompare = NOP

    visitConst = NOP

    visitContinue = NOP

    visitDecorators = NOP

    visitDict = NOP

    visitDiscard = NOP

    visitDiv = NOP

    visitEllipsis = NOP

    visitExec = NOP

    visitExpression = NOP

    visitFloorDiv = NOP

    visitFor = NOP

    def visitFrom(self, node):
        if self.importer is None:
            raise InspectError, "Please use the micropython.Importer class for code which uses the 'from' statement."

        module = self.importer.load(node.modname, 1)

        if module is None:
            print "Warning:", node.modname, "not imported."

        for name, alias in node.names:
            if name != "*":
                if module is not None and module.namespace.has_key(name):
                    self[alias or name] = attr = module[name]
                    if isinstance(attr, Module) and not attr.loaded:
                        self.importer.load(attr.name)
                else:
                    self[alias or name] = UnresolvedName(name, node.modname)
            else:
                if module is not None:
                    for n in module.namespace.keys():
                        self[n] = attr = module[n]
                        if isinstance(attr, Module) and not attr.loaded:
                            self.importer.load(attr.name)

        return None

    def visitFunction(self, node):
        function = Function(
            node.name,
            self.get_parent().full_name(),
            node.argnames,
            has_star=(node.flags & 4 != 0),
            has_dstar=(node.flags & 8 != 0)
            )

        self.namespaces.append(function)
        # Current namespace is the function.
        # Previous namespace is the class.
        if node.name == "__init__" and isinstance(self.namespaces[-2], Class):
            self.in_init = 1
        self.dispatch(node.code)
        self.in_init = 0
        self.namespaces.pop()

        self.store(node.name, function)
        return None

    visitGenExpr = NOP

    visitGenExprFor = NOP

    visitGenExprIf = NOP

    visitGenExprInner = NOP

    def visitGetattr(self, node):
        expr = self.dispatch(node.expr)
        if expr is not None:
            if isinstance(expr, Module):
                return expr.namespace.get(node.attrname)
            elif isinstance(expr, UnresolvedName):
                return UnresolvedName(node.attrname, expr.full_name())
        return builtins_namespace.get(node.attrname)

    def visitGlobal(self, node):
        if self.namespaces:
            for name in node.names:
                self.namespaces[-1].make_global(name)

    def visitIf(self, node):
        for test, body in node.tests:
            self.dispatch(body)
        if node.else_ is not None:
            self.dispatch(node.else_)
        return None

    visitIfExp = NOP

    def visitImport(self, node):
        if self.importer is None:
            raise InspectError, "Please use the micropython.Importer class for code which uses the 'import' statement."

        for name, alias in node.names:
            if alias is not None:
                self[alias] = self.importer.load(name, 1) or UnresolvedName(None, name)
            else:
                self[name.split(".")[0]] = self.importer.load(name) or UnresolvedName(None, name.split(".")[0])

        return None

    visitInvert = NOP

    visitKeyword = NOP

    visitLambda = NOP

    visitLeftShift = NOP

    visitList = NOP

    visitListComp = NOP

    visitListCompFor = NOP

    visitListCompIf = NOP

    visitMod = NOP

    def visitModule(self, node):
        return self.dispatch(node.node)

    visitMul = NOP

    def visitName(self, node):
        name = node.name
        if name == "self":
            return Self()
        elif self.has_key(name):
            return self[name]
        elif builtins_namespace.has_key(name):
            return builtins_namespace[name]
        else:
            return None

    visitNot = NOP

    visitOr = NOP

    visitPass = NOP

    visitPower = NOP

    visitPrint = NOP

    visitPrintnl = NOP

    visitRaise = NOP

    visitReturn = NOP

    visitRightShift = NOP

    visitSlice = NOP

    def visitStmt(self, node):
        for n in node.nodes:
            self.dispatch(n)
        return None

    visitSub = NOP

    visitSubscript = NOP

    def visitTryExcept(self, node):
        self.dispatch(node.body)
        for name, var, n in node.handlers:
            self.dispatch(n)
        if node.else_ is not None:
            self.dispatch(node.else_)
        return None

    visitTryFinally = NOP

    visitTuple = NOP

    visitUnaryAdd = NOP

    visitUnarySub = NOP

    visitWhile = NOP

    visitWith = NOP

    visitYield = NOP

class Self:

    "A reference to an object within a method."

    pass

# Built-in types initialisation.

builtins_namespace = {}
for key in ['ArithmeticError', 'AssertionError', 'AttributeError',
    'BaseException', 'DeprecationWarning', 'EOFError', 'Ellipsis',
    'EnvironmentError', 'Exception', 'False', 'FloatingPointError',
    'FutureWarning', 'GeneratorExit', 'IOError', 'ImportError',
    'ImportWarning', 'IndentationError', 'IndexError', 'KeyError',
    'KeyboardInterrupt', 'LookupError', 'MemoryError', 'NameError',
    'None', 'NotImplemented', 'NotImplementedError', 'OSError',
    'OverflowError', 'PendingDeprecationWarning', 'ReferenceError',
    'RuntimeError', 'RuntimeWarning', 'StandardError', 'StopIteration',
    'SyntaxError', 'SyntaxWarning', 'SystemError', 'SystemExit',
    'TabError', 'True', 'TypeError', 'UnboundLocalError',
    'UnicodeDecodeError', 'UnicodeEncodeError', 'UnicodeError',
    'UnicodeTranslateError', 'UnicodeWarning', 'UserWarning',
    'ValueError', 'Warning', 'ZeroDivisionError',
    'bool', 'buffer', 'complex', 'dict', 'file', 'float', 'int', 'list',
    'long', 'object', 'slice', 'str', 'tuple', 'type', 'unicode']:
    builtins_namespace[key] = Class(key, "__builtins__")

# vim: tabstop=4 expandtab shiftwidth=4
